import errno
import grp
import os
import pwd
import sqlite3
import shutil
import subprocess
import time

from charmhelpers import fetch
from charmhelpers.core import hookenv
from charmhelpers.core import host
from charmhelpers.core.services.base import service_restart
from charmhelpers.payload import (
    archive,
    execd,
    )


hooks = hookenv.Hooks()
config = hookenv.config()

# Globals
CHARM_FILES_DIR = os.path.join(hookenv.charm_dir(), 'files')
REQUIRED_PACKAGES = [
    'python-virtualenv',
    'python3-dev',
    'python3.4-venv',
    'libsqlite3-dev',
    'python3-jinja2',
    'python-swiftclient',
    ]

BASE_DIR = config['base_dir']
PAYLOADS_DIR = os.path.join(BASE_DIR, 'payloads')
CODE_DIR = os.path.join(BASE_DIR, 'code')
DB_DIR = os.path.join(BASE_DIR, 'db')
ENV_DIR = os.path.join(BASE_DIR, 'envs')  # environment variables
VENV_DIR = os.path.join(CODE_DIR, 'env')
LOGS_DIR = os.path.join(BASE_DIR, 'logs')
CODE_TARBALL = 'rutabaga.tar.gz'

CODE_USER = config['code_user']
CODE_GROUP = config['code_group']
USER = config['user']
GROUP = config['group']


def get_ports():
    return [s.strip() for s in config['ports'].split(',')]  # no array type


# Environment
def make_srv_location():
    hookenv.log('Creating directories...')

    for dir in (BASE_DIR, PAYLOADS_DIR):
        host.mkdir(dir, owner=CODE_USER, group=CODE_GROUP, perms=0o755)
    for dir in (LOGS_DIR, DB_DIR):
        host.mkdir(dir, owner=USER, group=GROUP, perms=0o755)


def execd_preinstall(service_name):
    execd.execd_preinstall()


def create_users(service_name):
    hookenv.log('Creating users and groups for %s.' % service_name)
    code_password = host.pwgen()
    host.adduser(CODE_USER, code_password)
    host.add_user_to_group(CODE_USER, CODE_GROUP)
    password = host.pwgen()
    host.adduser(USER, password)
    host.add_user_to_group(USER, GROUP)


def create_schema(service_name):
    hookenv.log('Creating database schema for %s.' % service_name)
    db_file = os.path.join(DB_DIR, 'rutabaga.db')
    db_schema = os.path.join(CODE_DIR, 'db', 'rutabaga.schema')
    if not os.path.isfile(db_file):
        conn = sqlite3.connect(db_file)
        with open(db_schema, 'r') as schema:
            conn.executescript(schema.read())
            conn.commit()
        conn.close()
    host.chownr(DB_DIR, USER, GROUP)


def get_swift_creds(config):
    return {
        'user': config['swift_username'],
        'project': config['swift_tenant_name'],
        'password': config['swift_password'],
        'authurl': config['swift_auth_url'],
        'region': config['swift_region_name'],
        }


def swift_base_cmd(**swift_creds):
    return [
        'swift',
        '--os-username=' + swift_creds['user'],
        '--os-tenant-name=' + swift_creds['project'],
        '--os-password=' + swift_creds['password'],
        '--os-auth-url=' + swift_creds['authurl'],
        '--os-region-name=' + swift_creds['region'],
        ]


def swift_get_etag(name, container=None, **swift_creds):
    cmd = swift_base_cmd(**swift_creds) + ['stat', container, name]
    file_stat = subprocess.check_output(cmd).splitlines()
    for line in file_stat:
        words = line.split()
        if words[0] == 'ETag:':
            return words[1]


def swift_fetch(source, target, container=None, **swift_creds):
    cmd = swift_base_cmd(**swift_creds) + [
        'download', '--output=' + target, container, source]
    subprocess.check_call(cmd)


def unlink_force(path):
    """Unlink path, without worrying about whether it exists."""
    try:
        os.unlink(path)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise


def symlink_force(source, link_name):
    """Create symlink link_name -> source, even if link_name exists."""
    unlink_force(link_name)
    os.symlink(source, link_name)


def install_python_packages(target_dir):
    hookenv.log('Installing Python dependencies...')
    # No `pyvenv` on trusty, so over-riding here
    subprocess.check_call(
        ['sudo', '-u', CODE_USER, 'PYVENV=pyvenv-3.4', 'make', '-C',
         target_dir, 'build',
         'PIP_SOURCE_DIR=%s' % os.path.join(target_dir, 'pip-cache')])


def prune_payloads(keep):
    for entry in os.listdir(PAYLOADS_DIR):
        if entry in keep:
            continue
        entry_path = os.path.join(PAYLOADS_DIR, entry)
        if os.path.isdir(entry_path):
            hookenv.log('Purging old build in %s...' % entry_path)
            shutil.rmtree(entry_path)


def deploy_code(service_name):
    make_srv_location()

    current_build_label = None
    if os.path.islink(CODE_DIR):
        current_build_label = os.path.basename(os.path.realpath(CODE_DIR))
    elif os.path.isdir(os.path.join(CODE_DIR, '.bzr')):
        log_output = subprocess.check_output(
            ['bzr', 'log', '-rlast:', '--show-ids', CODE_DIR])
        for line in log_output.splitlines():
            if line.startswith('revision-id: '):
                current_build_label = line[len('revision-id: '):]
    desired_build_label = config['build_label']
    if not desired_build_label:
        if current_build_label is not None:
            hookenv.log(
                'No desired build label, but build %s is already deployed' %
                current_build_label)
            return
        else:
            raise AssertionError('Build label unset, so cannot deploy code')
    if current_build_label == desired_build_label:
        hookenv.log('Build %s already deployed' % desired_build_label)
        return
    hookenv.log('Deploying build %s...' % desired_build_label)

    # Copy source archive
    archive_path = os.path.join(PAYLOADS_DIR, desired_build_label + '.tar.gz')
    object_name = os.path.join(desired_build_label, CODE_TARBALL)

    try:
        if config['swift_container_name']:
            swift_creds = get_swift_creds(config)
            swift_container = config['swift_container_name']
            swift_fetch(
                os.path.join('rutabaga-builds', object_name), archive_path,
                container=swift_container, **swift_creds)
        else:
            with open(os.path.join(CHARM_FILES_DIR, object_name)) as file:
                host.write_file(archive_path, file.read(), perms=0o644)

        # Unpack source
        target_dir = os.path.join(PAYLOADS_DIR, desired_build_label)
        if os.path.isdir(target_dir):
            shutil.rmtree(target_dir)
        archive.extract_tarfile(archive_path, target_dir)
        os.chown(
            target_dir,
            pwd.getpwnam(CODE_USER).pw_uid, grp.getgrnam(CODE_GROUP).gr_gid)
        host.lchownr(target_dir, CODE_USER, CODE_GROUP)

        install_python_packages(target_dir)

        if not os.path.islink(CODE_DIR) and os.path.isdir(CODE_DIR):
            old_payload_dir = os.path.join(PAYLOADS_DIR, current_build_label)
            if os.path.exists(old_payload_dir):
                shutil.rmtree(CODE_DIR)
            else:
                os.rename(CODE_DIR, old_payload_dir)
        symlink_force(
            os.path.relpath(target_dir, os.path.dirname(CODE_DIR)), CODE_DIR)
        prune_payloads([desired_build_label, current_build_label])
    finally:
        unlink_force(archive_path)


def install_packages(service_name):
    hookenv.log('Installing system packages...')
    fetch.configure_sources(update=True)
    fetch.apt_install(REQUIRED_PACKAGES, fatal=True)


def restart_rsync(service_name):
    # Only makes sense with basenode.
    if not config['log_hosts_allow']:
        return
    concat_rsync_fragments = '/usr/local/bin/concat_rsync_fragments'
    if not os.path.exists(concat_rsync_fragments):
        return
    subprocess.check_call([concat_rsync_fragments])
    service_restart('rsync')


def get_env_vars():
    # Get dict of environment variables from envdir directory.
    env_vars = {
        'PATH': '%s:$PATH' % os.path.join(VENV_DIR, 'bin'),
        'RUTABAGA_LOG_DIR': LOGS_DIR,
        }
    for file in os.listdir(ENV_DIR):
        with open(os.path.join(ENV_DIR, file), 'r') as f:
            env_vars[file] = f.read()
    return env_vars


def publish_wsgi_relations(service_name):
    # Publish the wsgi-file relation so the gunicorn subordinate can
    # serve us. Other WSGI containers could be made to work, as the most
    # gunicorn-specific thing is the --paste hack.
    config = hookenv.config()
    env = get_env_vars()
    for relid in hookenv.relation_ids('wsgi-file'):
        hookenv.relation_set(
            relid,
            working_dir=CODE_DIR,
            wsgi_wsgi_file='--paste rutabaga.ini',
            wsgi_user=config['user'],
            wsgi_group=config['group'],
            wsgi_log_file=os.path.join(LOGS_DIR, 'rutabaga.log'),
            wsgi_log_level='info',
            wsgi_access_logfile=os.path.join(LOGS_DIR, 'access.log'),
            wsgi_timestamp=time.time(),
            port=config['port_api'],
            env_extra=' '.join('%s=%s' % (k, v) for (k, v) in env.items()))


def nagios_restart(service_name):
    service_restart('nagios-nrpe-server')
